double expm1(double);
double log1p(double);
double acosh(double);
double asinh(double);
double atanh(double);
double rint(double);
double trunc(double);
double exp2(double);
double log2(double);
double hypot(double, double);
double atan2(double, double);
double pow(double, double);
double copysign(double, double);
double nextafter(double, double);
long long strtoll(const char*, char**, int);
unsigned long long strtoull(const char*, char**, int);
double cbrt(double);
long double sinl(long double);
long double cosl(long double);
long double tanl(long double);
long double sinhl(long double);
long double coshl(long double);
long double tanhl(long double);
long double fabsl(long double);
long double floorl(long double);
long double ceill(long double);
long double rintl(long double);
long double truncl(long double);
long double sqrtl(long double);
long double log10l(long double);
long double logl(long double);
long double log1pl(long double);
long double expl(long double);
long double expm1l(long double);
long double asinl(long double);
long double acosl(long double);
long double atanl(long double);
long double asinhl(long double);
long double acoshl(long double);
long double atanhl(long double);
long double hypotl(long double, long double);
long double atan2l(long double, long double);
long double powl(long double, long double);
long double fmodl(long double, long double);
long double modfl(long double, long double*);
long double frexpl(long double, int*);
long double ldexpl(long double, int);
long double exp2l(long double);
long double log2l(long double);
long double copysignl(long double, long double);
long double nextafterl(long double, long double);
long double cbrtl(long double);
float sinf(float);
float cosf(float);
float tanf(float);
float sinhf(float);
float coshf(float);
float tanhf(float);
float fabsf(float);
float floorf(float);
float ceilf(float);
float rintf(float);
float truncf(float);
float sqrtf(float);
float log10f(float);
float logf(float);
float log1pf(float);
float expf(float);
float expm1f(float);
float asinf(float);
float acosf(float);
float atanf(float);
float asinhf(float);
float acoshf(float);
float atanhf(float);
float hypotf(float, float);
float atan2f(float, float);
float powf(float, float);
float fmodf(float, float);
float modff(float, float*);
float frexpf(float, int*);
float ldexpf(float, int);
float exp2f(float);
float log2f(float);
float copysignf(float, float);
float nextafterf(float, float);
float cbrtf(float);
double sin(double);
double cos(double);
double tan(double);
double sinh(double);
double cosh(double);
double tanh(double);
double fabs(double);
double floor(double);
double ceil(double);
double sqrt(double);
double log10(double);
double log(double);
double exp(double);
double asin(double);
double acos(double);
double atan(double);
double fmod(double, double);
double modf(double, double*);
double frexp(double, int*);
double ldexp(double, int);
